# -*- coding: utf-8 -*-
import logging
import sys

from .constants import Logger

# define log formatter
log_format = "%(asctime)s [%(levelname)s] %(message)s"
formatter = logging.Formatter(log_format)


# define stream handler
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(formatter)


# define logger
logger = logging.getLogger(Logger.name)
logger.setLevel(logging.DEBUG)


# add handlers
logger.addHandler(stream_handler)



# -*- coding: utf-8 -*-
import logging
from typing import Dict, Tuple
from urllib.parse import urlencode
from uuid import uuid4

import httpx

from keycloak.config import config
from keycloak.constants import GrantTypes, Logger, ResponseTypes
from keycloak.utils import auth_header, handle_exceptions

log = logging.getLogger(Logger.name)


class AsyncAuthenticationMixin:
    """
    This class includes the methods to interact with the authentication flow
    """

    _userinfo: Dict = {}
    callback_uri = "http://localhost/kc/callback"

    async def login(self, scopes: Tuple = ("openid",)) -> Tuple:
        """
        method to generate openid login url and state

        >>> import uvicorn
        >>> from starlette.routing import Route
        >>> from starlette.applications import Starlette
        >>> from starlette.responses import RedirectResponse
        >>> from starlette.middleware import Middleware
        >>> from starlette.middleware.sessions import SessionMiddleware
        >>> from keycloak import AsyncClient
        >>>
        >>> kc= AsyncClient()
        >>>
        >>> async def howdy(request):
        >>>     return "Howdy!"
        >>>
        >>> async def login(request):
        >>>     url, state = await kc.login()
        >>>     request.session["state"] = state
        >>>     return RedirectResponse(url)
        >>>
        >>> routes = [
        >>>     Route("/howdy", howdy),
        >>>     Route("/login", login),
        >>> ]
        >>>
        >>> middlewares = [
        >>>     Middleware(SessionMiddleware, secret_key="bba085e8e95cd42cb19e268f")
        >>> ]
        >>>
        >>> app = Starlette(routes=routes, middleware=middlewares)
        >>>
        >>> if __name__ == "__main__":
        >>>     uvicorn.run(app)

        :param scopes: scopes to be requested eg: openid, email, profile etc

        :returns: endpoint url and state
        """
        state = uuid4().hex
        arguments = urlencode(
            {
                "state": state,
                "client_id": config.client.client_id,
                "response_type": ResponseTypes.code,
                "scope": " ".join(scopes),
                "redirect_uri": self.callback_uri,
            }
        )
        return f"{config.openid.authorization_endpoint}?{arguments}", state

    @handle_exceptions
    async def callback(self, code: str) -> Dict:
        """
        openid login callback handler

        >>> import uvicorn
        >>> from starlette.routing import Route
        >>> from starlette.applications import Starlette
        >>> from starlette.responses import RedirectResponse, PlainTextResponse
        >>> from starlette.middleware import Middleware
        >>> from starlette.middleware.sessions import SessionMiddleware
        >>> from keycloak import AsyncClient
        >>>
        >>> kc= AsyncClient()
        >>>
        >>> async def howdy(request):
        >>>     return "Howdy!"
        >>>
        >>> async def login(request):
        >>>     url, state = await kc.login()
        >>>     request.session["state"] = state
        >>>     return RedirectResponse(url)
        >>>
        >>> async def callback(request):
        >>>     state = request.query_params["state"]
        >>>     if request.session["state"] != state:
        >>>         return PlainTextResponse("Invalid state", status=400)
        >>>     code = request.query_params["code"]
        >>>     request.session["tokens"] = await kc.callback(code)
        >>>     return RedirectResponse("/howdy")
        >>>
        >>> routes = [
        >>>     Route("/howdy", howdy),
        >>>     Route("/login", login),
        >>>     Route("/callback", callback),
        >>> ]
        >>>
        >>> middlewares = [
        >>>     Middleware(SessionMiddleware, secret_key="bba085e8e95cd42cb19e268f")
        >>> ]
        >>>
        >>> app = Starlette(routes=routes, middleware=middlewares)
        >>>
        >>> if __name__ == "__main__":
        >>>     uvicorn.run(app)

        :param code: code send by the keycloak server
        :returns: dictionary
        """
        payload = {
            "code": code,
            "grant_type": GrantTypes.authorization_code,
            "redirect_uri": self.callback_uri,
            "client_id": config.client.client_id,
            "client_secret": config.client.client_secret,
        }
        log.debug("Retrieving user tokens from server")
        async with httpx.AsyncClient() as client:
            response = await client.post(config.openid.token_endpoint, data=payload)
            log.debug("User tokens retrieved successfully")
            return response.json()

    @handle_exceptions
    async def fetch_userinfo(self, access_token: str = None) -> Dict:
        """
        method to retrieve userinfo from the keycloak server

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(kc.fetch_userinfo())
        {'sub': '4c9c2430-b2e7-4f0b-9325-aa81dffe0463', 'email_verified': False, 'preferred_username': 'service-account-keycloak-client'}
        >>>

        :param access_token: access token of the client or user
        :returns: dicttionary
        """
        access_token = access_token or await self.access_token  # type: ignore
        headers = auth_header(access_token)
        log.debug("Retrieving user info from server")
        async with httpx.AsyncClient() as client:
            response = await client.post(
                config.openid.userinfo_endpoint, headers=headers
            )
            log.debug("User info retrieved successfully")
            return response.json()

    @property
    async def userinfo(self) -> Dict:
        """
        user information available within the server

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc = AsyncClient()
        >>> asyncio.run(kc.userinfo)
        {'sub': '4c9c2430-b2e7-4f0b-9325-aa81dffe0463', 'email_verified': False, 'preferred_username': 'service-account-keycloak-client'}
        >>>

        :returns: dictionary
        """
        if not self._userinfo:
            self._userinfo = await self.fetch_userinfo()
        return self._userinfo

    async def logout(self, access_token: str = None, refresh_token: str = None) -> None:
        access_token = access_token or await self.access_token  # type: ignore
        refresh_token = refresh_token or await self.refresh_token  # type: ignore
        payload = {
            "client_id": config.client.client_id,
            "client_secret": config.client.client_secret,
            "refresh_token": refresh_token,
        }
        headers = auth_header(access_token)
        log.debug("Logging out user from server")
        async with httpx.AsyncClient() as client:
            await client.post(
                config.openid.end_session_endpoint, data=payload, headers=headers
            )
            log.debug("User logged out successfully")



# -*- coding: utf-8 -*-
import base64
import json
import logging
from functools import wraps
from typing import Any, Callable, Dict, Tuple, Union

from requests.exceptions import HTTPError

from .constants import Headers, Logger, TokenType

log = logging.getLogger(Logger.name)


def b64encode(data: Any, serialize: bool = False) -> str:
    """method to encode string using base64"""
    serialized_data = json.dumps(data) if serialize else data
    data_as_bytes = serialized_data.encode("utf-8")
    return base64.b64encode(data_as_bytes).decode("utf-8")


def auth_header(token_val: str, token_type: str = TokenType.bearer) -> Dict:
    """method to generate authorization header"""
    return {Headers.authorization: f"{token_type} {token_val}"}


def basic_auth(username: str, password: str) -> Dict:
    """method to prepare the basic auth header"""
    token = f"{username}:{password}"
    token = b64encode(token)
    return auth_header(token, TokenType.basic)


def fix_padding(encoded_string: str) -> str:
    """method to correct padding for base64 encoding"""
    required_padding = len(encoded_string) % 4
    return encoded_string + ("=" * required_padding)


def handle_exceptions(func: Callable) -> Any:
    """decorator to take care of HTTPError"""

    @wraps(func)
    def wrapper(*args: Tuple, **kwargs: Dict) -> Any:
        try:
            return func(*args, **kwargs)
        except HTTPError as ex:
            if hasattr(ex.response, "content"):
                log.exception(ex.response.content)
            raise ex
        except Exception as ex:
            log.exception("Error occurred:")
            raise ex

    return wrapper


class Singleton(type):
    """metaclass for creating singleton classes"""

    _instances: Dict = {}

    def __call__(cls, *args: Tuple, **kwargs: Dict) -> type:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-
import logging
from typing import Dict, Tuple
from urllib.parse import urlencode
from uuid import uuid4

import httpx

from keycloak.config import config
from keycloak.constants import GrantTypes, Logger, ResponseTypes
from keycloak.utils import auth_header, handle_exceptions

log = logging.getLogger(Logger.name)


class AuthenticationMixin:
    """
    This class includes the methods to interact with the authentication flow
    """

    _userinfo: Dict = {}
    callback_uri = "http://localhost/kc/callback"

    def login(self, scopes: Tuple = ("openid",)) -> Tuple:
        """
        method to generate openid login url and state

        >>> from keycloak import Client
        >>> from flask import Flask, request, session, redirect
        >>>
        >>> kc = Client()
        >>>
        >>> app = Flask(__name__)
        >>>
        >>> @app.route("/howdy")
        >>> def howdy():
        >>>     return "Howdy!"
        >>>
        >>> @app.route("/login")
        >>> def login():
        >>>     url, state = kc.login()
        >>>     session["state"] = state
        >>>     return redirect(url)
        >>>
        >>> if __name__ == "__main__":
        >>>     app.run()

        :param scopes: scopes to be requested eg: openid, email, profile etc

        :returns: endpoint url and state
        """
        state = uuid4().hex
        arguments = urlencode(
            {
                "state": state,
                "client_id": config.client.client_id,
                "response_type": ResponseTypes.code,
                "scope": " ".join(scopes),
                "redirect_uri": self.callback_uri,
            }
        )
        return f"{config.openid.authorization_endpoint}?{arguments}", state

    @handle_exceptions
    def callback(self, code: str) -> Dict:
        """
        openid login callback handler

        >>> from keycloak import Client
        >>> from flask import Flask, request, session, redirect, Response
        >>>
        >>> kc = Client()
        >>>
        >>> app = Flask(__name__)
        >>>
        >>> @app.route("/howdy")
        >>> def howdy():
        >>>     return "Howdy!"
        >>>
        >>> @app.route("/login")
        >>> def login():
        >>>     url, state = kc.login()
        >>>     session["state"] = state
        >>>     return redirect(url)
        >>>
        >>> @app.route("/callback")
        >>> def callback():
        >>>     state = request.params["state"]
        >>>     if session["state"] != state:
        >>>         return Response("Invalid state", status=400)
        >>>
        >>>     code = request.params["code"]
        >>>     session["tokens"] = kc.callback(code)
        >>>     return redirect("/howdy")
        >>>
        >>> if __name__ == "__main__":
        >>>     app.run()

        :param code: code send by the keycloak server
        :returns: dictionary
        """
        payload = {
            "code": code,
            "grant_type": GrantTypes.authorization_code,
            "redirect_uri": self.callback_uri,
            "client_id": config.client.client_id,
            "client_secret": config.client.client_secret,
        }
        log.debug("Retrieving user tokens from server")
        response = httpx.post(config.openid.token_endpoint, data=payload)
        response.raise_for_status()
        log.debug("User tokens retrieved successfully")
        return response.json()

    @handle_exceptions
    def fetch_userinfo(self, access_token: str = None) -> Dict:
        """
        method to retrieve userinfo from the keycloak server

        >>>
        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.fetch_userinfo()
        {'sub': '4c9c2430-b2e7-4f0b-9325-aa81dffe0463', 'email_verified': False, 'preferred_username': 'service-account-keycloak-client'}
        >>>

        :param access_token: access token of the client or user
        :returns: dicttionary
        """
        access_token = access_token or self.access_token  # type: ignore
        headers = auth_header(access_token)
        log.debug("Retrieving user info from server")
        response = httpx.post(config.openid.userinfo_endpoint, headers=headers)
        response.raise_for_status()
        log.debug("User info retrieved successfully")
        return response.json()

    @property
    def userinfo(self) -> Dict:
        """
        user information available within the server

        >>>
        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.userinfo
        {'sub': '4c9c2430-b2e7-4f0b-9325-aa81dffe0463', 'email_verified': False, 'preferred_username': 'service-account-keycloak-client'}
        >>>

        :returns: dictionary
        """
        if not self._userinfo:
            self._userinfo = self.fetch_userinfo()
        return self._userinfo

    def logout(self, access_token: str = None, refresh_token: str = None) -> None:
        access_token = access_token or self.access_token  # type: ignore
        refresh_token = refresh_token or self.refresh_token  # type: ignore
        payload = {
            "client_id": config.client.client_id,
            "client_secret": config.client.client_secret,
            "refresh_token": refresh_token,
        }
        headers = auth_header(access_token)
        log.debug("Logging out user from server")
        response = httpx.post(
            config.openid.end_session_endpoint, data=payload, headers=headers
        )
        response.raise_for_status()
        log.debug("User logged out successfully")



# -*- coding: utf-8 -*-
import json
from typing import Any, Callable, Dict

from flask import Config, Flask, Request, redirect
from flask.sessions import SessionInterface
from werkzeug.wrappers import Response

from .. import Client


class ProxyApp:
    def __init__(self, config: Config):
        self.config = config

    @property
    def secret_key(self) -> str:
        return self.config["SECRET_KEY"]

    @property
    def session_cookie_name(self) -> str:
        return self.config["SESSION_COOKIE_NAME"]

    @property
    def permanent_session_lifetime(self) -> int:
        return self.config["PERMANENT_SESSION_LIFETIME"]  # pragma: nocover


class AuthenticationMiddleware:
    session_interface: SessionInterface = None  # type: ignore

    def __init__(
        self,
        app: Flask,
        config: Config,
        session_interface: SessionInterface,
        callback_url: str = "http://localhost:5000/kc/callback",
        login_redirect_uri: str = "/",
        logout_uri: str = "/kc/logout",
        logout_redirect_uri: str = "/",
    ) -> None:
        self.app = app
        self.config = config
        self.session_interface = session_interface
        self.callback_url = callback_url
        self.login_redirect_uri = login_redirect_uri
        self.logout_uri = logout_uri
        self.logout_redirect_uri = logout_redirect_uri
        self.kc = Client(callback_url)
        self.proxy_app = ProxyApp(config)

    def _response(
        self, environ: Dict, start_response: Callable, session: Any, response: Callable
    ) -> Callable:
        self.session_interface.save_session(self.proxy_app, session, response)
        return response(environ, start_response)

    def __call__(self, environ: Dict, start_response: Callable) -> Callable:
        response = None
        request = Request(environ)
        session = self.session_interface.open_session(  # type: ignore
            self.proxy_app, request
        )

        # callback request
        if request.base_url == self.callback_url:
            response = self.callback(session, request)
            return self._response(environ, start_response, session, response)

        # logout request
        elif request.path == self.logout_uri:
            response = self.logout(session)
            return self._response(environ, start_response, session, response)

        # logout redirect uri
        elif request.path == self.logout_redirect_uri:
            return self.app(environ, start_response)

        # unauthorized request
        elif "user" not in session:
            response = self.login(session)
            return self._response(environ, start_response, session, response)

        # authorized request
        else:
            return self.app(environ, start_response)

    def login(self, session: Dict) -> Response:
        """Initiate authentication"""
        url, state = self.kc.login()
        session["state"] = state
        return redirect(url)

    def callback(self, session: Dict, request: Request) -> Response:
        """Authentication callback handler"""

        # validate state
        state = request.args.get("state", "unknown")
        _state = session.pop("state", None)
        if state != _state:
            return Response("Invalid state", status=403)

        # fetch user tokens
        code: str = request.args.get("code", "unknown")
        tokens = self.kc.callback(code)
        session["tokens"] = json.dumps(tokens)

        # fetch user info
        access_token = tokens["access_token"]
        user = self.kc.fetch_userinfo(access_token)
        session["user"] = json.dumps(user)

        return redirect(self.login_redirect_uri)

    def logout(self, session: Dict) -> Response:

        if "tokens" in session:
            tokens = json.loads(session["tokens"])
            access_token = tokens["access_token"]
            refresh_token = tokens["refresh_token"]
            self.kc.logout(access_token, refresh_token)
            del session["tokens"]

        if "user" in session:
            del session["user"]

        return redirect(self.logout_redirect_uri)



# -*- coding: utf-8 -*-
import logging
from typing import Dict, List

import httpx
from cached_property import cached_property
from jose import jwt

from keycloak.config import OpenId, config
from keycloak.constants import Logger
from keycloak.utils import basic_auth, handle_exceptions

log = logging.getLogger(Logger.name)


class AsyncTokenMixin:
    """This class consists of methods that can be user to perform JWT operations"""

    _tokens: Dict = {}
    openid: OpenId = None  # type: ignore

    @property
    async def tokens(self) -> Dict:
        """
        access and refresh tokens associated with the client/user

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.tokens)
        {'access_token': 'eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJLRG9qblhUaF90Z0dzeWtDM1g4VjJfaEY3TU0zZlZpa1B6T2VDX2RiLWx3In0.eyJqdGkiOiIzOWJkMDQ0ZS04OGQ3LTQ3MGQtOGE4Ni1hMWMxMWI2ZmI2MzIiLCJleHAiOjE1ODQyNTcwMjgsIm5iZiI6MCwiaWF0IjoxNTg0MjU2OTY4LCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwiYXVkIjpbImtleWNsb2FrLWNsaWVudCIsImFjY291bnQiXSwic3ViIjoiNmJhMDg0NDMtMzg4MS00MGU3LWFmNDMtMTBiNzE5NmIwMmZkIiwidHlwIjoiQmVhcmVyIiwiYXpwIjoia2V5Y2xvYWstY2xpZW50IiwiYXV0aF90aW1lIjowLCJzZXNzaW9uX3N0YXRlIjoiNWU3ZDcyMjUtMGVjNi00NTgxLWE3OWEtMDkyMWI2ZjYzMDFhIiwiYWNyIjoiMSIsInJlYWxtX2FjY2VzcyI6eyJyb2xlcyI6WyJvZmZsaW5lX2FjY2VzcyIsInVtYV9hdXRob3JpemF0aW9uIl19LCJyZXNvdXJjZV9hY2Nlc3MiOnsia2V5Y2xvYWstY2xpZW50Ijp7InJvbGVzIjpbInVtYV9wcm90ZWN0aW9uIl19LCJhY2NvdW50Ijp7InJvbGVzIjpbIm1hbmFnZS1hY2NvdW50IiwibWFuYWdlLWFjY291bnQtbGlua3MiLCJ2aWV3LXByb2ZpbGUiXX19LCJzY29wZSI6ImVtYWlsIHByb2ZpbGUiLCJlbWFpbF92ZXJpZmllZCI6ZmFsc2UsImNsaWVudElkIjoia2V5Y2xvYWstY2xpZW50IiwiY2xpZW50SG9zdCI6IjE3Mi4xNy4wLjEiLCJwcmVmZXJyZWRfdXNlcm5hbWUiOiJzZXJ2aWNlLWFjY291bnQta2V5Y2xvYWstY2xpZW50IiwiY2xpZW50QWRkcmVzcyI6IjE3Mi4xNy4wLjEifQ.Yx_GlS3VUl9SfqH185A1G-ZKnTLdPVkRaLoNlXZYG-yA3g85293T4cRayqWDy-aSs0QrhcFht4E-x0PfXZXkS-F_mdByJXzkojCtS3zwztCvUt1667RTc-tRqxqiU3jzdrQw8pQSPKJBBwwrVOthoowc17uSCr7QmvNf2VUA6411ySyzrqr2Hm5Y8oUpcY-eAeN3ivxuIHfCIyv2QRj4MchthY8mZkZGltJygZIN3fHSqiNmgv29NGQDZQ7JMDRg1dKDqFVEmmhObtmjiYu1jpxX3Z3rLGw3AfpdzkfLlC7Bd0qNjLIYQDfmniX1-LK9eEzvUGSD3Mf0bHvfgw5qlg', 'expires_in': 60, 'refresh_expires_in': 1800, 'refresh_token': 'eyJhbGciOiJIUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJjYmFkZWRhYS04NDA5LTQ3MDYtODcwMy02NjViM2Q2MzYwMDAifQ.eyJqdGkiOiJkODVhOWJlYi03NzE1LTQ4ZTAtOTA3My00ZDEyMmQ4MDM1NzAiLCJleHAiOjE1ODQyNTg3NjgsIm5iZiI6MCwiaWF0IjoxNTg0MjU2OTY4LCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwiYXVkIjoiaHR0cDovL2xvY2FsaG9zdDo4MDgwL2F1dGgvcmVhbG1zL21hc3RlciIsInN1YiI6IjZiYTA4NDQzLTM4ODEtNDBlNy1hZjQzLTEwYjcxOTZiMDJmZCIsInR5cCI6IlJlZnJlc2giLCJhenAiOiJrZXljbG9hay1jbGllbnQiLCJhdXRoX3RpbWUiOjAsInNlc3Npb25fc3RhdGUiOiI1ZTdkNzIyNS0wZWM2LTQ1ODEtYTc5YS0wOTIxYjZmNjMwMWEiLCJyZWFsbV9hY2Nlc3MiOnsicm9sZXMiOlsib2ZmbGluZV9hY2Nlc3MiLCJ1bWFfYXV0aG9yaXphdGlvbiJdfSwicmVzb3VyY2VfYWNjZXNzIjp7ImtleWNsb2FrLWNsaWVudCI6eyJyb2xlcyI6WyJ1bWFfcHJvdGVjdGlvbiJdfSwiYWNjb3VudCI6eyJyb2xlcyI6WyJtYW5hZ2UtYWNjb3VudCIsIm1hbmFnZS1hY2NvdW50LWxpbmtzIiwidmlldy1wcm9maWxlIl19fSwic2NvcGUiOiJlbWFpbCBwcm9maWxlIn0.K22gl2ICSmVh6omZMk7z4DB_Io4tDiWp__Gxxx_W4Nk', 'token_type': 'bearer', 'not-before-policy': 0, 'session_state': '5e7d7225-0ec6-4581-a79a-0921b6f6301a', 'scope': 'email profile'}
        >>>

        :returns: dictionary
        """
        if not self._tokens:
            self._tokens = await self.pat(self.username, self.password)  # type: ignore
        return self._tokens

    @tokens.setter
    async def tokens(self, val: Dict) -> None:
        """setter for tokens"""
        self._tokens = val

    @property
    async def access_token(self) -> str:
        """
        access token associated with the client/user

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.access_token)
        'eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJLRG9qblhUaF90Z0dzeWtDM1g4VjJfaEY3TU0zZlZpa1B6T2VDX2RiLWx3In0.eyJqdGkiOiIzOWJkMDQ0ZS04OGQ3LTQ3MGQtOGE4Ni1hMWMxMWI2ZmI2MzIiLCJleHAiOjE1ODQyNTcwMjgsIm5iZiI6MCwiaWF0IjoxNTg0MjU2OTY4LCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwiYXVkIjpbImtleWNsb2FrLWNsaWVudCIsImFjY291bnQiXSwic3ViIjoiNmJhMDg0NDMtMzg4MS00MGU3LWFmNDMtMTBiNzE5NmIwMmZkIiwidHlwIjoiQmVhcmVyIiwiYXpwIjoia2V5Y2xvYWstY2xpZW50IiwiYXV0aF90aW1lIjowLCJzZXNzaW9uX3N0YXRlIjoiNWU3ZDcyMjUtMGVjNi00NTgxLWE3OWEtMDkyMWI2ZjYzMDFhIiwiYWNyIjoiMSIsInJlYWxtX2FjY2VzcyI6eyJyb2xlcyI6WyJvZmZsaW5lX2FjY2VzcyIsInVtYV9hdXRob3JpemF0aW9uIl19LCJyZXNvdXJjZV9hY2Nlc3MiOnsia2V5Y2xvYWstY2xpZW50Ijp7InJvbGVzIjpbInVtYV9wcm90ZWN0aW9uIl19LCJhY2NvdW50Ijp7InJvbGVzIjpbIm1hbmFnZS1hY2NvdW50IiwibWFuYWdlLWFjY291bnQtbGlua3MiLCJ2aWV3LXByb2ZpbGUiXX19LCJzY29wZSI6ImVtYWlsIHByb2ZpbGUiLCJlbWFpbF92ZXJpZmllZCI6ZmFsc2UsImNsaWVudElkIjoia2V5Y2xvYWstY2xpZW50IiwiY2xpZW50SG9zdCI6IjE3Mi4xNy4wLjEiLCJwcmVmZXJyZWRfdXNlcm5hbWUiOiJzZXJ2aWNlLWFjY291bnQta2V5Y2xvYWstY2xpZW50IiwiY2xpZW50QWRkcmVzcyI6IjE3Mi4xNy4wLjEifQ.Yx_GlS3VUl9SfqH185A1G-ZKnTLdPVkRaLoNlXZYG-yA3g85293T4cRayqWDy-aSs0QrhcFht4E-x0PfXZXkS-F_mdByJXzkojCtS3zwztCvUt1667RTc-tRqxqiU3jzdrQw8pQSPKJBBwwrVOthoowc17uSCr7QmvNf2VUA6411ySyzrqr2Hm5Y8oUpcY-eAeN3ivxuIHfCIyv2QRj4MchthY8mZkZGltJygZIN3fHSqiNmgv29NGQDZQ7JMDRg1dKDqFVEmmhObtmjiYu1jpxX3Z3rLGw3AfpdzkfLlC7Bd0qNjLIYQDfmniX1-LK9eEzvUGSD3Mf0bHvfgw5qlg'
        >>>

        :returns: string
        """
        tokens = await self.tokens
        return tokens["access_token"]

    @property
    async def refresh_token(self) -> str:
        """
        refresh token associated with the client/user

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(kc.refresh_token)
        'eyJhbGciOiJIUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJjYmFkZWRhYS04NDA5LTQ3MDYtODcwMy02NjViM2Q2MzYwMDAifQ.eyJqdGkiOiJkODVhOWJlYi03NzE1LTQ4ZTAtOTA3My00ZDEyMmQ4MDM1NzAiLCJleHAiOjE1ODQyNTg3NjgsIm5iZiI6MCwiaWF0IjoxNTg0MjU2OTY4LCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwiYXVkIjoiaHR0cDovL2xvY2FsaG9zdDo4MDgwL2F1dGgvcmVhbG1zL21hc3RlciIsInN1YiI6IjZiYTA4NDQzLTM4ODEtNDBlNy1hZjQzLTEwYjcxOTZiMDJmZCIsInR5cCI6IlJlZnJlc2giLCJhenAiOiJrZXljbG9hay1jbGllbnQiLCJhdXRoX3RpbWUiOjAsInNlc3Npb25fc3RhdGUiOiI1ZTdkNzIyNS0wZWM2LTQ1ODEtYTc5YS0wOTIxYjZmNjMwMWEiLCJyZWFsbV9hY2Nlc3MiOnsicm9sZXMiOlsib2ZmbGluZV9hY2Nlc3MiLCJ1bWFfYXV0aG9yaXphdGlvbiJdfSwicmVzb3VyY2VfYWNjZXNzIjp7ImtleWNsb2FrLWNsaWVudCI6eyJyb2xlcyI6WyJ1bWFfcHJvdGVjdGlvbiJdfSwiYWNjb3VudCI6eyJyb2xlcyI6WyJtYW5hZ2UtYWNjb3VudCIsIm1hbmFnZS1hY2NvdW50LWxpbmtzIiwidmlldy1wcm9maWxlIl19fSwic2NvcGUiOiJlbWFpbCBwcm9maWxlIn0.K22gl2ICSmVh6omZMk7z4DB_Io4tDiWp__Gxxx_W4Nk'
        >>>

        :returns: string
        """
        tokens = await self.tokens
        return tokens["refresh_token"]

    @property
    async def scope(self) -> str:
        """
        scopes available

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.scope)
        ['email', 'profile']
        >>>

        :returns: list
        """
        tokens = await self.tokens
        return tokens["scope"].split(" ")

    @property
    async def token_type(self) -> str:
        """
        type of token

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.token_type)
        'bearer'
        >>>

        :returns: string
        """
        tokens = await self.tokens
        return tokens["token_type"]

    @handle_exceptions
    async def refresh_tokens(self) -> None:
        """
        method to refresh expired access token using refresh token

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.refresh_tokens())
        """
        headers = basic_auth(config.client.client_id, config.client.client_secret)
        payload = {
            "client_id": config.client.client_id,
            "grant_type": "refresh_token",
            "refresh_token": self._tokens["refresh_token"],
        }
        log.debug("Refreshing tokens")
        async with httpx.AsyncClient() as client:
            response = await client.post(
                config.uma2.token_endpoint, data=payload, headers=headers
            )
            log.debug("Tokens refreshed successfully")
            self._tokens = response.json()

    @cached_property
    async def jwks(self) -> List:
        """
        list of signing keys/JWKs used by the keycloak server

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(kc.jwks)
        [{'kid': 'KDojnXTh_tgGsykC3X8V2_hF7MM3fVikPzOeC_db-lw', 'kty': 'RSA', 'alg': 'RS256', 'use': 'sig', 'n': 'q8JxMYdetQKGZHhH6ZzQvnc0S6qfBlUchwuNITAD2_nBlta970_2zE840bxbwQSiZVfMh1fdnQ4xZiIc5qTjeLIn2n6LBs78uzTdAP4PG1tyV2jJviBnJY6FNEHwWKJ-bPLMp_WRze5uSnzwW9sq2e9XhhQY1os9m6tou01GIo93KUnYY94Xvl1MMNjxFAX7RA5MYi-qPw6BNi-b_5WB1LD3A2e-aJUnh40NMwJAPC286Is8KvJIgeg3CMIKPfvVcwzMDUvrLZHRHhvypjvW5ws7OgnNkngCdnz8hyMe4qGNkKl8rbWiu5UOI6qlG3ub4r_CTv6nvqIOdlyO_y97wQ', 'e': 'AQAB', 'x5c': ['MIICmzCCAYMCBgFw3RFEWTANBgkqhkiG9w0BAQsFADARMQ8wDQYDVQQDDAZtYXN0ZXIwHhcNMjAwMzE1MDcxOTIxWhcNMzAwMzE1MDcyMTAxWjARMQ8wDQYDVQQDDAZtYXN0ZXIwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCrwnExh161AoZkeEfpnNC+dzRLqp8GVRyHC40hMAPb+cGW1r3vT/bMTzjRvFvBBKJlV8yHV92dDjFmIhzmpON4sifafosGzvy7NN0A/g8bW3JXaMm+IGcljoU0QfBYon5s8syn9ZHN7m5KfPBb2yrZ71eGFBjWiz2bq2i7TUYij3cpSdhj3he+XUww2PEUBftEDkxiL6o/DoE2L5v/lYHUsPcDZ75olSeHjQ0zAkA8Lbzoizwq8kiB6DcIwgo9+9VzDMwNS+stkdEeG/KmO9bnCzs6Cc2SeAJ2fPyHIx7ioY2QqXyttaK7lQ4jqqUbe5viv8JO/qe+og52XI7/L3vBAgMBAAEwDQYJKoZIhvcNAQELBQADggEBAEIkR4iXNmFWJTyP5j682i7MWeHnMCTMQTQZtOfqbVAJLKMRxm9X8ND/PwbucoLFsmx9UnyCe9erIJ/EGRimTNGij311kRWVtDCw6FtObMaxNvgOLwiuN989N7bA2B6QqsBqYvlYnEHy5rPLJWRojUPQ942mildqRiFI5JfGHjY88gwju/q/DhhAPu7wFORYDYxo0Fxv4/aV/VJzg03gav0/vZWebqEa/aRFSpScZgYxc4KloNMkYZYHYo/OtAX01WUBO5cilbennNnUDeryy3FBc9p1/rv1a1BL9rY5wW2Kt4jYyp82lk+gZqtfHJDHHk44QLW7xq37AYd8JdQMjss='], 'x5t': 'eDziHrpDesOJXALdEuBF2tY4vWc', 'x5t#S256': '__cmdM6AqteNxXRMH4daJGEsG15jZCsSxKTCwYk-PX4'}]
        >>>

        :returns: list
        """
        log.debug("Fectching JWK keys")
        async with httpx.AsyncClient() as client:
            response = await client.get(config.openid.jwks_uri)
            data = response.json()
            return data["keys"]

    async def decode(self, token: str) -> Dict:
        """
        decode given json web token (jwt)

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.decode(await kc.access_token))
        {'jti': '39bd044e-88d7-470d-8a86-a1c11b6fb632', 'exp': 1584257028, 'nbf': 0, 'iat': 1584256968, 'iss': 'http://localhost:8080/auth/realms/master', 'aud': ['keycloak-client', 'account'], 'sub': '6ba08443-3881-40e7-af43-10b7196b02fd', 'typ': 'Bearer', 'azp': 'keycloak-client', 'auth_time': 0, 'session_state': '5e7d7225-0ec6-4581-a79a-0921b6f6301a', 'acr': '1', 'realm_access': {'roles': ['offline_access', 'uma_authorization']}, 'resource_access': {'keycloak-client': {'roles': ['uma_protection']}, 'account': {'roles': ['manage-account', 'manage-account-links', 'view-profile']}}, 'scope': 'email profile', 'email_verified': False, 'clientId': 'keycloak-client', 'clientHost': '172.17.0.1', 'preferred_username': 'service-account-keycloak-client', 'clientAddress': '172.17.0.1'}
        >>>

        :param token: jwt to be decoded eg:access_token or refresh_token
        :returns: dictionary
        """
        return jwt.decode(
            token,
            await self.jwks,  # type: ignore
            issuer=config.openid.issuer,
            audience=config.client.client_id,
        )



# -*- coding: utf-8 -*-
import json
from typing import Callable

from django.conf import settings
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect

from keycloak import Client


class AuthenticationMiddleware:
    def __init__(self, get_response: Callable):
        self.get_response = get_response
        self.kc = Client(self.callback_url)

    @property
    def login_redirect_uri(self) -> str:
        return settings.KEYCLOAK_LOGIN_REDIRECT_URI

    @property
    def logout_redirect_uri(self) -> str:
        return settings.KEYCLOAK_LOGOUT_REDIRECT_URI

    @property
    def logout_uri(self) -> str:
        if hasattr(settings, "KEYCLOAK_LOGOUT_URI"):
            return settings.KEYCLOAK_LOGOUT_URI
        else:
            return "/kc/logout"

    @property
    def callback_url(self) -> str:
        return settings.KEYCLOAK_CALLBACK_URL

    def callback(self, request: HttpRequest) -> HttpResponse:

        # validate state
        state = request.GET.get("state", "unknown")
        _state = request.session.pop("state", None)
        if state != _state:
            return HttpResponse("Invalid state", status=403)

        # fetch user tokens
        code: str = request.GET.get("code", "unknown")
        tokens = self.kc.callback(code)
        request.session["tokens"] = json.dumps(tokens)

        # fetch user info
        access_token = tokens["access_token"]
        user = self.kc.fetch_userinfo(access_token)
        request.session["user"] = json.dumps(user)

        return HttpResponseRedirect(self.login_redirect_uri)

    def login(self, request: HttpRequest) -> HttpResponse:
        """Initiate authentication"""
        url, state = self.kc.login()
        request.session["state"] = state
        return HttpResponseRedirect(url)

    def logout(self, request: HttpRequest) -> HttpResponse:

        if "tokens" in request.session:
            tokens = json.loads(request.session["tokens"])
            access_token = tokens["access_token"]
            refresh_token = tokens["refresh_token"]
            self.kc.logout(access_token, refresh_token)
            del request.session["tokens"]

        if "user" in request.session:
            del request.session["user"]

        return HttpResponseRedirect(self.logout_redirect_uri)

    def __call__(self, request: HttpRequest) -> HttpResponse:

        # callback request
        if request.build_absolute_uri(request.path) == self.callback_url:
            return self.callback(request)

        # logout request
        elif request.path == self.logout_uri:
            return self.logout(request)

        # logout redirect uri
        elif request.path == self.logout_redirect_uri:
            return self.get_response(request)

        # unauthorized request
        elif "user" not in request.session:
            return self.login(request)

        # authorized request
        else:
            return self.get_response(request)



# -*- coding: utf-8 -*-
import logging
from typing import Dict, List

import httpx

from keycloak.config import config
from keycloak.constants import Logger
from keycloak.utils import auth_header, handle_exceptions

log = logging.getLogger(Logger.name)


class AsyncResourceMixin:
    """
    This class consists of methods that can be used to manage resources
    """

    _resources: List = []

    @property
    async def resources(self) -> List:
        """
        list of resources available in keycloak server

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.resources)
        [{'name': 'default Resource', 'type': 'urn:python-client:resources:default', 'owner': {'id': 'd74cc555-d46c-4ef8-8a30-ceb2b91d8823'}, 'ownerManagedAccess': False, 'attributes': {}, '_id': 'bb6a777f-a17b-4555-b035-a6ce12a1fd21', 'uris': ['/*'], 'resource_scopes': []}]
        >>>

        :returns: list
        """
        if not self._resources:
            self._resources = await self.find_resources()
        return self._resources

    @handle_exceptions
    async def find_resources(self, access_token: str = None) -> List:
        """
        fetch resources from keycloak server

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.find_resources())
        [{'name': 'default Resource', 'type': 'urn:python-client:resources:default', 'owner': {'id': 'd74cc555-d46c-4ef8-8a30-ceb2b91d8823'}, 'ownerManagedAccess': False, 'attributes': {}, '_id': 'bb6a777f-a17b-4555-b035-a6ce12a1fd21', 'uris': ['/*'], 'resource_scopes': []}]
        >>>

        :param access_token: access token to be used
        :returns: list
        """
        access_token = access_token or await self.access_token  # type: ignore
        headers = auth_header(access_token)
        log.debug("Retrieving resources from keycloak")
        async with httpx.AsyncClient() as client:
            response = await client.get(config.uma2.resource_endpoint, headers=headers)
            log.debug("Resources retrieved successfully")
            return [await self.find_resource(x, access_token) for x in response.json()]  # type: ignore

    @handle_exceptions
    async def find_resource(self, resource_id: str, access_token: str = None) -> Dict:
        """
        Method to fetch the details of a resource

        >>> import asyncio
        >>> from keycloak import AsyncClient
        >>> kc= AsyncClient()
        >>> asyncio.run(await kc.find_resource('bb6a777f-a17b-4555-b035-a6ce12a1fd21'))
        {'name': 'default Resource', 'type': 'urn:python-client:resources:default', 'owner': {'id': 'd74cc555-d46c-4ef8-8a30-ceb2b91d8823'}, 'ownerManagedAccess': False, 'attributes': {}, '_id': 'bb6a777f-a17b-4555-b035-a6ce12a1fd21', 'uris': ['/*'], 'resource_scopes': []}
        >>>

        :param resource_id: id of the resource
        :param access_token: access token to be used
        :returns: list
        """
        access_token = access_token or await self.access_token  # type: ignore
        headers = auth_header(access_token)
        endpoint = f"{config.uma2.resource_endpoint}/{resource_id}"
        log.debug("Retrieving resource from keycloak")
        async with httpx.AsyncClient() as client:
            response = await client.get(endpoint, headers=headers)
            log.debug("Resource retrieved successfully")
            return response.json()



# -*- coding: utf-8 -*-
import logging
from dataclasses import asdict
from typing import Dict, List, Tuple

import httpx

from keycloak.config import config
from keycloak.constants import GrantTypes, Logger, TokenType, TokenTypeHints
from keycloak.utils import auth_header, basic_auth, handle_exceptions

log = logging.getLogger(Logger.name)


class AuthorizationMixin:
    """
    collection of methods to interact with the authorization api
    see https://www.keycloak.org/docs/latest/authorization_services/ for details
    """

    _ticket: Dict = None  # type: ignore
    _rpt: Dict = {}

    @staticmethod
    def payload_for_client() -> Dict:
        """
        generate payload for keycloak client

        :returns: dictionary
        """
        return {"grant_type": GrantTypes.client_credentials}

    @staticmethod
    def payload_for_user(username: str = None, password: str = None) -> Dict:
        """
        generate payload for keycloak user

        :param username: username to be used
        :param password: password to be used

        :returns: dictionary
        """
        if username and password:
            return {
                "grant_type": GrantTypes.password,
                "username": username,
                "password": password,
            }
        else:
            return {}

    @staticmethod
    @handle_exceptions
    def pat(username: str = None, password: str = None) -> Dict:
        """
        retrieve protection api token (PAT),
        see `docs <https://www.keycloak.org/docs/latest/authorization_services/#_service_protection_whatis_obtain_pat>`__ for more details

        >>>
        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.pat()
        {'access_token': 'eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJmMTNTUndXVGxRQnFmNmNoamM4SHRtY09sN3NpaTNpaGVXaFNLX0hla280In0.eyJqdGkiOiIzNDliZmQ1ZC1jZDZjLTRjZDAtOTM3Mi1lNzE2N2Y5NGEyMDkiLCJleHAiOjE1ODQxODU0MDcsIm5iZiI6MCwiaWF0IjoxNTg0MTg1MzQ3LCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwiYXVkIjpbImtleWNsb2FrLWNsaWVudCIsImFjY291bnQiXSwic3ViIjoiNGM5YzI0MzAtYjJlNy00ZjBiLTkzMjUtYWE4MWRmZmUwNDYzIiwidHlwIjoiQmVhcmVyIiwiYXpwIjoia2V5Y2xvYWstY2xpZW50IiwiYXV0aF90aW1lIjowLCJzZXNzaW9uX3N0YXRlIjoiYWY3N2JmOTEtNTQ2ZC00M2E1LTgwZmYtYzQ0MDg2MDVhNTlmIiwiYWNyIjoiMSIsInJlYWxtX2FjY2VzcyI6eyJyb2xlcyI6WyJvZmZsaW5lX2FjY2VzcyIsInVtYV9hdXRob3JpemF0aW9uIl19LCJyZXNvdXJjZV9hY2Nlc3MiOnsia2V5Y2xvYWstY2xpZW50Ijp7InJvbGVzIjpbInVtYV9wcm90ZWN0aW9uIl19LCJhY2NvdW50Ijp7InJvbGVzIjpbIm1hbmFnZS1hY2NvdW50IiwibWFuYWdlLWFjY291bnQtbGlua3MiLCJ2aWV3LXByb2ZpbGUiXX19LCJzY29wZSI6ImVtYWlsIHByb2ZpbGUiLCJjbGllbnRJZCI6ImtleWNsb2FrLWNsaWVudCIsImVtYWlsX3ZlcmlmaWVkIjpmYWxzZSwiY2xpZW50SG9zdCI6IjE3Mi4xNy4wLjEiLCJwcmVmZXJyZWRfdXNlcm5hbWUiOiJzZXJ2aWNlLWFjY291bnQta2V5Y2xvYWstY2xpZW50IiwiY2xpZW50QWRkcmVzcyI6IjE3Mi4xNy4wLjEifQ.PUq6-N-9Hn57kSrK0-6reC8AYT08EV4DZsBqxQ7D0cyp4hHTSP5ax9zQwn0Q05bCZ3c_L_prwMa8VOEFF5Nv-ch2otVg7lIqTc4xeLORSRmIxhjFSxLbiXwBhl4mmfOOHwSL0yBRVHhPzdRkHyVmLlq1WrPfiNe6NRuhDiQIIaFTA0-jSfyqtXNcYjcidsTwjL0q9_Cxt0BrsjTQZLO2b9NKnh5rKbBewvwwIWvZqY5j8QiT3LL652RAED_q7RbG-jFqtLleEklKTtS3bFlJIMCzvdOmfTxZT-FY2zbx_0NJil8bQTJorwveF6Vjb11vAuTI9ccjaOg54SFQT3tSGQ', 'expires_in': 60, 'refresh_expires_in': 1800, 'refresh_token': 'eyJhbGciOiJIUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICI5MmE5NTI2NC0wNTAyLTQzN2ItYWE3ZS01ZGIwNjFlYzMwOWYifQ.eyJqdGkiOiJhYjU2NTg1OS1hODJkLTQzMzYtOTFmNS1kNTU0OWU2ZTU0YTkiLCJleHAiOjE1ODQxODcxNDcsIm5iZiI6MCwiaWF0IjoxNTg0MTg1MzQ3LCJpc3MiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwiYXVkIjoiaHR0cDovL2xvY2FsaG9zdDo4MDgwL2F1dGgvcmVhbG1zL21hc3RlciIsInN1YiI6IjRjOWMyNDMwLWIyZTctNGYwYi05MzI1LWFhODFkZmZlMDQ2MyIsInR5cCI6IlJlZnJlc2giLCJhenAiOiJrZXljbG9hay1jbGllbnQiLCJhdXRoX3RpbWUiOjAsInNlc3Npb25fc3RhdGUiOiJhZjc3YmY5MS01NDZkLTQzYTUtODBmZi1jNDQwODYwNWE1OWYiLCJyZWFsbV9hY2Nlc3MiOnsicm9sZXMiOlsib2ZmbGluZV9hY2Nlc3MiLCJ1bWFfYXV0aG9yaXphdGlvbiJdfSwicmVzb3VyY2VfYWNjZXNzIjp7ImtleWNsb2FrLWNsaWVudCI6eyJyb2xlcyI6WyJ1bWFfcHJvdGVjdGlvbiJdfSwiYWNjb3VudCI6eyJyb2xlcyI6WyJtYW5hZ2UtYWNjb3VudCIsIm1hbmFnZS1hY2NvdW50LWxpbmtzIiwidmlldy1wcm9maWxlIl19fSwic2NvcGUiOiJlbWFpbCBwcm9maWxlIn0.vdL4dCB8GLa-045EO3_UNUfDWMTOvoDnFs9giRjrvQM', 'token_type': 'bearer', 'not-before-policy': 0, 'session_state': 'af77bf91-546d-43a5-80ff-c4408605a59f', 'scope': 'email profile'}
        >>>

        :param username: username to be used
        :param password: password to be used

        :returns: dictionary
        """
        headers = basic_auth(config.client.client_id, config.client.client_secret)
        payload = (
            AuthorizationMixin.payload_for_user(username, password)
            or AuthorizationMixin.payload_for_client()
        )
        log.debug("Retrieving PAT from server")
        response = httpx.post(config.uma2.token_endpoint, data=payload, headers=headers)
        response.raise_for_status()
        return response.json()

    @handle_exceptions
    def ticket(self, resources: List = [], access_token: str = None) -> Dict:
        """
        retrieve permission ticket from keycloak server
        see `docs <https://www.keycloak.org/docs/latest/authorization_services/#_overview_terminology_permission_ticket>`__ for more details

        >>>
        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.ticket(kc.resources)
        {'ticket': 'eyJhbGciOiJIUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICI5MmE5NTI2NC0wNTAyLTQzN2ItYWE3ZS01ZGIwNjFlYzMwOWYifQ.eyJwZXJtaXNzaW9ucyI6W3sicnNpZCI6IjQ4MTUyMTI2LWFhOTEtNGE0Zi04ZWU4LTczOTI4ZjViNmMwMCJ9XSwianRpIjoiNDhlZmVmZDQtMDg5NC00NjA2LTk0YjUtMDhlNjZiNTE4YWM2LTE1ODQxODUyOTUyMzkiLCJleHAiOjE1ODQxODUzNTUsIm5iZiI6MCwiaWF0IjoxNTg0MTg1Mjk1LCJhdWQiOiJodHRwOi8vbG9jYWxob3N0OjgwODAvYXV0aC9yZWFsbXMvbWFzdGVyIiwic3ViIjoiNGM5YzI0MzAtYjJlNy00ZjBiLTkzMjUtYWE4MWRmZmUwNDYzIiwiYXpwIjoia2V5Y2xvYWstY2xpZW50In0.Or_6wzK9wlQMBPpi8bWIioWCeO6QuolKjr4mKC4YWpA'}
        >>>

        :param resources: list of resources
        :param access_token: access token to be used

        :returns: dictionary
        """
        access_token = access_token or self.access_token  # type: ignore
        resources = resources or self.resources  # type: ignore
        headers = auth_header(access_token, TokenType.bearer)
        payload = [
            {"resource_id": x["_id"], "resource_scopes": x["resource_scopes"]}
            for x in resources
        ]
        log.debug("Retrieving permission ticket from keycloak")
        response = httpx.post(
            config.uma2.permission_endpoint, json=payload, headers=headers
        )
        response.raise_for_status()
        log.debug("Permission ticket retrieved successfully")
        return response.json()

    @handle_exceptions
    def rpt(self, access_token: str) -> Dict:
        """
        retrieve request party token (RPT)
        see `docs <https://www.keycloak.org/docs/latest/authorization_services/#_service_rpt_overview>`__ for more details

        >>>
        >>> form keycloak import Client
        >>> kc = Client(username='myuser', password='*****')
        >>> kc.rpt(kc.access_token)
        2020-08-03 11:47:54,568 [DEBUG] Retrieving RPT from keycloak
        2020-08-03 11:47:54,581 [DEBUG] RPT retrieved successfully
        2020-08-03 11:47:54,581 [DEBUG] Retrieving JWKs from keycloak server
        2020-08-03 11:47:54,587 [DEBUG] JWKs retrieved successfully
        {'exp': 1596435534, 'iat': 1596435474, 'jti': '822c7f6f-cd0a-4b9d-b55c-72803755ca7c', 'iss': 'http://localhost:8080/auth/realms/master', 'aud': ['kc', 'account'], 'sub': 'fce7f440-2e40-4161-90b4-61a1913c4b28', 'typ': 'Bearer', 'azp': 'kc', 'session_state': 'b3498962-3973-4f2d-9150-b1338e974d08', 'acr': '1', 'realm_access': {'roles': ['offline_access', 'uma_authorization']}, 'resource_access': {'account': {'roles': ['manage-account', 'manage-account-links', 'view-profile']}}, 'authorization': {'permissions': [{'rsid': '992dfa45-6098-45ac-b62e-da4d2787377f', 'rsname': 'Default Resource'}]}, 'scope': 'profile email', 'email_verified': True, 'preferred_username': 'cute-as-duck'}
        >>>

        :param access_token: access token to be used

        :returns: dictionary
        """
        payload = {
            "grant_type": GrantTypes.uma_ticket,
            "audience": config.client.client_id,
        }
        headers = auth_header(access_token, TokenType.bearer)
        log.debug("Retrieving RPT from keycloak")
        response = httpx.post(config.uma2.token_endpoint, data=payload, headers=headers)
        response.raise_for_status()
        log.debug("RPT retrieved successfully")
        return response.json()

    @staticmethod
    @handle_exceptions
    def introspect(rpt: str) -> Dict:
        """
        introspect the request party token (RPT)
        see `docs <https://www.keycloak.org/docs/latest/authorization_services/#_service_protection_token_introspection>`__ for more details

        >>>
         >>> form keycloak import Client
        >>> kc = Client(username='myuser', password='*****')
        >>> rpt = kc.rpt(kc.access_token)
        >>> kc.introspect(rpt['access_token'])
        2020-08-03 11:47:54,628 [DEBUG] Introspecting RPT token
        2020-08-03 11:47:54,635 [DEBUG] RPT introspected successfully
        {'exp': 1596435534, 'nbf': 0, 'iat': 1596435474, 'jti': '822c7f6f-cd0a-4b9d-b55c-72803755ca7c', 'aud': ['kc', 'account'], 'typ': 'Bearer', 'acr': '1', 'permissions': [{'rsid': '992dfa45-6098-45ac-b62e-da4d2787377f', 'rsname': 'Default Resource', 'resource_id': '992dfa45-6098-45ac-b62e-da4d2787377f', 'resource_scopes': []}], 'active': True}
        >>>

        :param rpt: rpt token

        :returns: dictionary
        """
        payload = {"token_type_hint": TokenTypeHints.rpt, "token": rpt}
        headers = basic_auth(config.client.client_id, config.client.client_secret)
        log.debug("Introspecting RPT token")
        response = httpx.post(
            config.uma2.introspection_endpoint, data=payload, headers=headers
        )
        response.raise_for_status()
        log.debug("RPT introspected successfully")
        return response.json()



# -*- coding: utf-8 -*-

from keycloak.core.asynchronous.authentication import AsyncAuthenticationMixin
from keycloak.core.asynchronous.authorization import AsyncAuthorizationMixin
from keycloak.core.asynchronous.resource import AsyncResourceMixin
from keycloak.core.asynchronous.token import AsyncTokenMixin
from keycloak.core.authentication import AuthenticationMixin
from keycloak.core.authorization import AuthorizationMixin
from keycloak.core.resource import ResourceMixin
from keycloak.core.token import TokenMixin
from keycloak.utils import Singleton


class Client(
    AuthenticationMixin,
    AuthorizationMixin,
    TokenMixin,
    ResourceMixin,
    metaclass=Singleton,
):
    """
    Python client to interact with the rest APIs provided by the keycloak server
    """

    def __init__(
        self,
        callback_uri: str = "http://localhost/kc/callback",
        username: str = None,
        password: str = None,
    ) -> None:
        self.callback_uri = callback_uri
        self.username = username
        self.password = password


class AsyncClient(
    AsyncAuthenticationMixin,
    AsyncAuthorizationMixin,
    AsyncTokenMixin,
    AsyncResourceMixin,
    metaclass=Singleton,
):
    def __init__(
        self,
        callback_uri: str = "http://localhost/kc/callback",
        username: str = None,
        password: str = None,
    ) -> None:
        self.callback_uri = callback_uri
        self.username = username
        self.password = password


__all__ = ["Client", "AsyncClient"]



# -*- coding: utf-8 -*-


class Logger:
    """constants associated with logger"""

    name = "keycloak"


class Defaults:
    """constants associated with default values"""

    keycloak_settings = "keycloak.json"


class EnvVar:
    """constants associated with env vars"""

    keycloak_settings = "KEYCLOAK_SETTINGS"


class FileMode:
    """constants associated with file mode"""

    read_only = "r"


class TokenType:
    """constants associated with token types"""

    basic = "Basic"
    bearer = "Bearer"


class Headers:
    """constants associated with headers"""

    authorization = "Authorization"


class GrantTypes:
    """constants associated with grant types"""

    password = "password"
    authorization_code = "authorization_code"
    client_credentials = "client_credentials"
    uma_ticket = "urn:ietf:params:oauth:grant-type:uma-ticket"


class ResponseTypes:
    """constants associated with response types"""

    code = "code"


class TokenTypeHints:
    """constants associated with token type hints"""

    rpt = "requesting_party_token"


class Algorithms:
    """constants associated with algorithms"""

    ec = ("ES256", "ES384", "ES521", "ES512")
    hmac = ("HS256", "HS384", "HS512")
    rsa = ("RS256", "RS384", "RS512")
    rsapss = ("PS256", "PS384", "PS512")



# -*- coding: utf-8 -*-
import logging
from dataclasses import dataclass
from typing import Dict, List

import httpx

from keycloak.config import config
from keycloak.constants import Logger
from keycloak.utils import auth_header, handle_exceptions

log = logging.getLogger(Logger.name)


class ResourceMixin:
    """
    This class consists of methods that can be used to manage resources
    """

    _resources: List = []

    @property
    def resources(self) -> List:
        """
        list of resources available in keycloak server

        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.resources
        [{'name': 'Default Resource', 'type': 'urn:python-client:resources:default', 'owner': {'id': 'd74cc555-d46c-4ef8-8a30-ceb2b91d8823'}, 'ownerManagedAccess': False, 'attributes': {}, '_id': 'bb6a777f-a17b-4555-b035-a6ce12a1fd21', 'uris': ['/*'], 'resource_scopes': []}]
        >>>

        :returns: list
        """
        if not self._resources:
            self._resources = self.find_resources()
        return self._resources

    @handle_exceptions
    def find_resources(self, access_token: str = None) -> Dict:
        """
        fetch resources from keycloak server

        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.find_resources()
        [{'name': 'Default Resource', 'type': 'urn:python-client:resources:default', 'owner': {'id': 'd74cc555-d46c-4ef8-8a30-ceb2b91d8823'}, 'ownerManagedAccess': False, 'attributes': {}, '_id': 'bb6a777f-a17b-4555-b035-a6ce12a1fd21', 'uris': ['/*'], 'resource_scopes': []}]
        >>>

        :param access_token: access token to be used
        :returns: list
        """
        access_token = access_token or self.access_token  # type: ignore
        headers = auth_header(access_token)
        log.debug("Retrieving resources from keycloak")
        response = httpx.get(config.uma2.resource_endpoint, headers=headers)
        response.raise_for_status()
        log.debug("Resources retrieved successfully")
        return [self.find_resource(x, access_token) for x in response.json()]  # type: ignore

    @handle_exceptions
    def find_resource(self, resource_id: str, access_token: str = None) -> Dict:
        """
        Method to fetch the details of a resource

        >>> from keycloak import Client
        >>> kc = Client()
        >>> kc.find_resource('bb6a777f-a17b-4555-b035-a6ce12a1fd21')
        {'name': 'Default Resource', 'type': 'urn:python-client:resources:default', 'owner': {'id': 'd74cc555-d46c-4ef8-8a30-ceb2b91d8823'}, 'ownerManagedAccess': False, 'attributes': {}, '_id': 'bb6a777f-a17b-4555-b035-a6ce12a1fd21', 'uris': ['/*'], 'resource_scopes': []}
        >>>

        :param access_token: access token to be used
        :returns: list
        """
        access_token = access_token or self.access_token  # type: ignore
        headers = auth_header(access_token)
        endpoint = f"{config.uma2.resource_endpoint}/{resource_id}"
        log.debug("Retrieving resource from keycloak")
        response = httpx.get(endpoint, headers=headers)
        response.raise_for_status()
        log.debug("Resource retrieved successfully")
        return response.json()



# -*- coding: utf-8 -*-
import json
import logging
import os
from dataclasses import dataclass, fields
from typing import Dict, List

import requests
from cached_property import cached_property

from .constants import Defaults, EnvVar, FileMode, Logger
from .utils import Singleton

log = logging.getLogger(Logger.name)


class DataClassMixin:
    def __init__(self, **kwargs: Dict):
        attrs = [x.name for x in fields(self)]
        for key, val in kwargs.items():
            key = key.replace("-", "_")
            if key in attrs:
                setattr(self, key, val)


@dataclass(init=False)
class Client(DataClassMixin):
    realm: str
    auth_server_url: str
    ssl_required: str
    resource: str
    verify_token_audience: str
    credentials: Dict
    confidential_port: int
    policy_enforcer: Dict

    @property
    def client_id(self) -> str:
        return self.resource

    @property
    def client_secret(self) -> str:
        return self.credentials["secret"]


@dataclass(init=False)
class OpenId(DataClassMixin):
    issuer: str
    authorization_endpoint: str
    token_endpoint: str
    userinfo_endpoint: str
    end_session_endpoint: str
    jwks_uri: str
    introspection_endpoint: str


@dataclass(init=False)
class Uma2(DataClassMixin):
    issuer: str
    authorization_endpoint: str
    token_endpoint: str
    end_session_endpoint: str
    jwks_uri: str
    resource_registration_endpoint: str
    permission_endpoint: str
    policy_endpoint: str
    introspection_endpoint: str

    @property
    def resource_endpoint(self) -> str:
        return self.resource_registration_endpoint


class Config(metaclass=Singleton):
    @property
    def settings_file(self) -> str:
        log.debug("Lookup settings file in the env vars")
        return os.getenv(EnvVar.keycloak_settings, Defaults.keycloak_settings)

    @cached_property
    def client(self) -> Client:
        log.debug("Loading client config from the settings file")
        with open(self.settings_file, FileMode.read_only) as stream:
            data = json.loads(stream.read())
            return Client(**data)

    @property
    def openid_endpoint(self) -> str:
        auth_server_url = self.client.auth_server_url.rstrip("/")
        return (
            auth_server_url
            + "/realms/"
            + self.client.realm
            + "/.well-known/openid-configuration"
        )

    @cached_property
    def openid(self) -> OpenId:
        log.debug("Loading openid config using well-known endpoint")
        response = requests.get(self.openid_endpoint)
        response.raise_for_status()
        data = response.json()
        return OpenId(**data)

    @property
    def uma_endpoint(self) -> str:
        auth_server_url = self.client.auth_server_url.rstrip("/")
        return (
            auth_server_url
            + "/realms/"
            + self.client.realm
            + "/.well-known/uma2-configuration"
        )

    @cached_property
    def uma2(self) -> Uma2:
        log.debug("Loading uma2 config using well-known endpoint")
        response = requests.get(self.uma_endpoint)
        response.raise_for_status()
        data = response.json()
        return Uma2(**data)


config: Config = Config()



# -*- coding: utf-8 -*-
import json
from typing import Any

from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import PlainTextResponse, RedirectResponse, Response
from starlette.types import ASGIApp, Receive, Scope, Send

from keycloak import AsyncClient


class EndpointHandler(HTTPEndpoint):
    def __init__(self, *args: Any, **kwargs: Any):
        self.kc = kwargs.pop("kc")
        self.redirect_uri = kwargs.pop("redirect_uri")
        super().__init__(*args, **kwargs)


class Login(EndpointHandler):
    async def get(self, request: Request) -> Response:
        url, state = await self.kc.login()
        request.session["state"] = state
        return RedirectResponse(url)


class Logout(EndpointHandler):
    async def get(self, request: Request) -> Response:
        # retrieve tokens
        _tokens = request.session["tokens"]
        tokens = json.loads(_tokens)
        access_token = tokens["access_token"]
        refresh_token = tokens["refresh_token"]

        # logout from keycloak
        await self.kc.logout(access_token, refresh_token)
        del request.session["tokens"]

        # delete user info
        if "user" in request.session:
            del request.session["user"]

        return RedirectResponse(self.redirect_uri)


class Callback(EndpointHandler):
    async def get(self, request: Request) -> Response:

        # validate state
        state = request.query_params["state"]
        _state = request.session.pop("state", "unknown")
        if state != _state:
            return PlainTextResponse("Invalid state", status_code=403)

        # retrieve tokens
        code = request.query_params["code"]
        tokens = await self.kc.callback(code)

        # starlette sessions do not have the capacity to store the entire tokens
        # so storing only access token and refresh token
        _tokens = {
            "access_token": tokens["access_token"],
            "refresh_token": tokens["refresh_token"],
        }
        request.session["tokens"] = json.dumps(_tokens)

        # retrieve user info
        access_token = tokens["access_token"]
        user = await self.kc.fetch_userinfo(access_token)
        request.session["user"] = json.dumps(user)

        return RedirectResponse(self.redirect_uri)


class AuthenticationMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        callback_url: str = "https://localhost:8000/kc/callback",
        login_redirect_uri: str = "/",
        logout_uri: str = "/kc/logout",
        logout_redirect_uri: str = "/",
    ) -> None:
        self.app = app
        self.callback_url = callback_url
        self.login_redirect_uri = login_redirect_uri
        self.logout_uri = logout_uri
        self.logout_redirect_uri = logout_redirect_uri
        self.kc = AsyncClient(callback_url)

    @staticmethod
    def get_url(request: Request) -> str:
        return f"{request.url.scheme}://{request.url.netloc}{request.url.path}"

    @staticmethod
    def is_http(scope: Any) -> bool:
        return scope["type"] == "http"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:

        # handle http requests
        if self.is_http(scope):
            request = Request(scope, receive)

            # handle callback request
            if self.get_url(request) == self.callback_url:
                await Callback(
                    scope,
                    receive,
                    send,
                    kc=self.kc,
                    redirect_uri=self.login_redirect_uri,
                )
                return

            # handle logout request
            elif request.url.path == self.logout_uri:
                await Logout(
                    scope,
                    receive,
                    send,
                    kc=self.kc,
                    redirect_uri=self.logout_redirect_uri,
                )
                return

            # handle logout redirect uri
            elif request.url.path == self.logout_redirect_uri:
                await self.app(scope, receive, send)
                return

            # handle unauthorized requests
            elif "user" not in request.session:
                await Login(
                    scope,
                    receive,
                    send,
                    kc=self.kc,
                    redirect_uri=self.logout_redirect_uri,
                )
                return

            # handle authorized requests
            else:
                await self.app(scope, receive, send)
                return

        # handle non http requests
        await self.app(scope, receive, send)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['keycloak',
 'keycloak.core',
 'keycloak.core.asynchronous',
 'keycloak.extensions']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'httpx>=0.18.2,<0.19.0',
 'python-jose[cryptography]>=3.3.0,<4.0.0']

extras_require = \
{'docs': ['Sphinx>=4.0.2,<5.0.0', 'sphinx-rtd-theme>=0.5.2,<0.6.0'],
 'extensions': ['Flask>=2.0.1,<3.0.0',
                'starlette>=0.15.0,<0.16.0',
                'Django>=3.2.4,<4.0.0',
                'uvicorn>=0.14.0,<0.15.0']}

setup_kwargs = {
    'name': 'keycloak',
    'version': '3.1.2',
    'description': 'Python client for Keycloak IAM',
    'long_description': None,
    'author': 'Cute As Duck',
    'author_email': 'pronoun_pears_0m@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
